library nui_push;

import 'package:flutter/cupertino.dart';
import 'package:nui_core/code/nui_device_util.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_push/model/nui_push_callback.dart';
import 'package:nui_push/model/nui_push_models.dart';
import 'package:nui_web/nui_web.dart';
import 'package:onesignal_flutter/onesignal_flutter.dart';

export 'package:nui_push/model/nui_push_callback.dart';
export 'package:nui_push/model/nui_push_models.dart';

abstract class NUIPushInterface{
  Future<DataResult> registerPushDeviceForUser({required String deviceToken});
  Future<DataResult> updateReadStatus({required String recipientId});
  Future<DataResult> setPushDeviceMeta({required String userId, required Map<String, dynamic> params});
  Future<DataResult> changeUserId(String userId);
  Future setLocationShared(bool enable);
  Future<DataResult> deletePushDeviceMeta({required String userId, required List<String> params});
  Future<DataResult<List<NUIPushMessage>>> getPushMessages({required int page, required int limit});
  Future<DataResult<List<NUIPushInboxMessage>>> getInboxNotifications({required int page, required int limit});
  Future<DataResult<int>> getInboxNotificationCount();
  Future<DataResult> setInboxNotificationRead({List<String>? notificationId, bool isRead = true});
  Future<DataResult<NUIPushMessage>> getPushMessageById({required String recipientId});
  Future<Map<String, dynamic>> getTags();
  Future sendTag(String key, dynamic value);
  Future<Map<String, dynamic>> sendTags(Map<String, dynamic> tags);
  Future deleteTag(String key);
  Future deleteTags(List<String> keys);
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToUsers({required String title, String? subtitle, required String content, NUIEnt? data, List<String>? userIds, String? bigPictureUrl, String? channelId});
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToAll({required String title, String? subtitle, required String content, NUIEnt? data, List<String>? includedSegments, List<String>? excludedSegments, String? bigPictureUrl, String? channelId});
}

class NUIPush implements NUIPushInterface{
  static const String _MODULE = "NUIPush";
  static const String NUIWEB_MODULE = "NUIPushWebModule";
  static const String _NUIWEB_ONESIGNAL = "NUIPushWebOSModule";
  static const String _ONE_SIGNAL_REST_URL = "https://onesignal.com/api/v1/notifications";
  static const String _OS_USER_ID = "NUIPushUserID";

  static NUIPush? _instance;
  static NUIPush? get instance {
    if(_instance == null){
      throwNUI("NUIPush", "NUIPush has not been built with NUIPushBuilder");
    }
    return _instance;
  }

  NUIPush._builder(NUIPushBuilder builder){
    _oneSignalAuthKey = builder._oneSignalAuthKey;
    _oneSignalAppId = builder._oneSignalAppId;
    _userId = builder._userId;
    _notificationReceivedListener = builder._notificationReceivedListener;
    _notificationOpenedListener = builder._notificationOpenedListener;
    _instance = this;
  }

  late String _userId;
  String? _oneSignalAuthKey;
  late String _oneSignalAppId;
  NUIPushNotificationListener? _notificationReceivedListener;
  NUIPushNotificationListener? _notificationOpenedListener;

  @override
  Future<DataResult<NUIPushMessage>> getPushMessageById({required String recipientId}) async{
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doGET<NUIPushMessage>(url: "push/my-messages", pathVariable: recipientId, entity: NUIPushMessageEntity());
    if(result.success == true){
      logNUI(_MODULE, "Get push message of recipient id : $recipientId successful");
      return DataResult<NUIPushMessage>(success: true, message: result.message, data: result.body);
    }
    else{
      logNUI(_MODULE, "Failed to get push message of recipient id : $recipientId");
      return DataResult(success: false, message: result.message);
    }
  }

  void _saveUserID(String userId) async{
    await NUISharedPref.get().save(_OS_USER_ID, userId);
  }

  Future<String> getUserID() async{
    final deviceState = await OneSignal.shared.getDeviceState();
    final userId = deviceState?.userId;
    return userId ?? await NUISharedPref.get().getString(_OS_USER_ID, defValue: "");
  }

  void _registerUserIDIfChanged(String deviceToken) async{
    final prevUserId = await NUISharedPref.get().getString(_OS_USER_ID);
    if(!match(prevUserId, deviceToken)){
      registerPushDeviceForUser(deviceToken: deviceToken);
    }
  }

  @override
  Future setLocationShared(bool enable) async {
    final hasPermission = await NUIPermissionUtil.hasLocationPermission();
    if(hasPermission) {
      final result = await OneSignal.shared.setLocationShared(enable);
    }
  }

  @override
  Future<DataResult<List<NUIPushMessage>>> getPushMessages({required int page, required int limit}) async{
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doGET(url: "push/my-messages", params: {
      "page": page.toString(),
      "limit": limit.toString()
    });

    if(result.success == true){
      logNUI(_MODULE, "Get push messages successful");
      return DataResult<List<NUIPushMessage>>(success: true, message: result.message, data: asListOfObjects(NUIPushMessageEntity(), result.responseBody));
    }
    else{
      logNUI(_MODULE, "Failed to get push messages");
      return DataResult(success: false, message: result.message);
    }
  }

  Future<DataResult> registerPushWithUser(String userId) async{
    final osUserId = await getUserID();
    final deviceInfo = await NUIDeviceUtil.getDeviceInfo();
    final registerRequest = NUIRegisterPushDeviceRequest(userId: userId, uniqueId: deviceInfo.uuid ?? "", providerDeviceToken: osUserId, deviceOs: deviceInfo.platform, deviceModel: deviceInfo.model);

    final result = await NUIWeb.get(module: NUIWEB_MODULE).doPOST<NUIStandardHttpResponse>(url: "push/registerNew", bodyEnt: registerRequest, entity: NUIStandardHttpResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Register device successful");
      return DataResult(success: true, message: "Register device successful");
    }
    else{
      logNUI(_MODULE, "Failed to register device");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult> changeUserId(String userId, {bool reRegister:false}) async{
    final newUserId = userId;
    final oldUserId = _userId;
    this._userId = userId;
    if(!match(newUserId, oldUserId)){
      return registerPushDeviceForUser(deviceToken: await getUserID());
    }
    else{
      if(reRegister == true){
        return registerPushDeviceForUser(deviceToken: await getUserID());
      }
      return DataResult(success: true, message: "User ID changed with the same value");
    }
  }

  @override
  Future<DataResult> registerPushDeviceForUser({required String deviceToken}) async{
    if(isNullOrEmpty(deviceToken)) return DataResult(success: false, message: "Failed to register device due to missing device token");

    final deviceInfo = await NUIDeviceUtil.getDeviceInfo();
    final registerRequest = NUIRegisterPushDeviceRequest(userId: _userId, uniqueId: deviceInfo.uuid ?? "", providerDeviceToken: deviceToken, deviceOs: deviceInfo.platform, deviceModel: deviceInfo.model);

    final result = await NUIWeb.get(module: NUIWEB_MODULE).doPOST<NUIStandardHttpResponse>(url: "push/registerNew", bodyEnt: registerRequest, entity: NUIStandardHttpResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Register device successful");
      return DataResult(success: true, message: "Register device successful");
    }
    else{
      logNUI(_MODULE, "Failed to register device");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult> setPushDeviceMeta({required String userId, required Map<String, dynamic> params}) async{
    if(isNullOrEmpty(userId)) return DataResult(success: false, message: "Invalid user ID: $userId");
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doPOST<NUIStandardHttpResponse>(url: "push/devices/$userId/user-meta", body: params, entity: NUIStandardHttpResponseEntity());
    final updateResult = await sendTags(params);
    if(result.success == true){
      logNUI(_MODULE, "Update device push metadata successful");
      return DataResult(success: true, message: "Update device push metadata successful");
    }
    else{
      logNUI(_MODULE, "Failed to update device push metadata");
      return DataResult(success: false, message: result.message);
    }
  }

  Future<Map<String, dynamic>> getTags() async{
    return OneSignal.shared.getTags();
  }

  Future sendTag(String key, dynamic value) async{
    await OneSignal.shared.sendTag(key, value);
  }

  Future<Map<String, dynamic>> sendTags(Map<String, dynamic> tags) async{
    return OneSignal.shared.sendTags(tags);
  }

  Future deleteTag(String key) async{
    await OneSignal.shared.deleteTag(key);
  }

  Future deleteTags(List<String> keys) async{
    await OneSignal.shared.deleteTags(keys);
  }

  @override
  Future<DataResult> deletePushDeviceMeta({required String userId, required List<String> params}) async{
    if(isNullOrEmpty(userId)) return DataResult(success: false, message: "Invalid user ID: $userId");
    final result = deleteTags(params);
    logNUI(_MODULE, "Delete device push metadatas successful");
    return DataResult(success: true, message: "Update device push metadata successful");
  }

  @override
  Future<DataResult> updateReadStatus({required String recipientId}) async{
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doPUT<NUIStandardHttpResponse>(url: "push/message/read", pathVariable: recipientId, entity: NUIStandardHttpResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Update message read status successful");
      return DataResult(success: true, message: "Update message read status successful");
    }
    else{
      logNUI(_MODULE, "Failed to update message read status");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToAll({required String title, required String content, String? subtitle, NUIEnt? data,  List<String>? includedSegments, List<String>? excludedSegments, String? bigPictureUrl, String? channelId}) async{
    try{
      logNUI(_MODULE, "Sending push notification to all");
      final notification = NUIPushOSNotification(
        appId: _oneSignalAppId,
        headings: {"en": title},
        subtitle: subtitle != null ? {"en": subtitle} : null,
        contents: {"en": content},
        data: data?.toMap(),
        bigPicture: bigPictureUrl,
        huawei_big_picture: bigPictureUrl,
        huawei_channel_id: channelId,
        android_channel_id: channelId,
        includedSegments: includedSegments ?? [],
        excludedSegments: excludedSegments ?? [],
      );

      final result = await NUIWeb.get(module: _NUIWEB_ONESIGNAL).doPOST<NUIPushOSResponse>(bodyEnt: notification, headers: _oneSignalPushHeaders(), entity: NUIPushOSResponseEntity());
      if(result.success == true){
        logNUI(_MODULE, "Send push message to all successful");
        return DataResult(success: true, message: "Send push message to all successful");
      }
      else{
        logNUI(_MODULE, "Failed to send push message to all with message: ${result.message}");
        return DataResult(success: false, message: "Failed to send push message to all with message : ${result.message}");
      }
    }catch(e){
      logNUI(_MODULE, "Failed to send push message to all with error : $e");
      return DataResult(success: false, message: "Failed to send push message to all with error: $e");
    }
  }

  Map<String, String> _oneSignalPushHeaders(){
    return {
      "Content-Type": "application/json; charset=UTF-8",
      "Authorization": "Basic $_oneSignalAuthKey"
    };
  }

  @override
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToUsers({required String title, String? subtitle, required String content, NUIEnt? data, List<String>? userIds, String? bigPictureUrl, String? channelId}) async{
    try{
      logNUI(_MODULE, "Sending push notification to users : $userIds");
      final notification = NUIPushOSNotification(
        appId: _oneSignalAppId,
        headings: {"en": title},
        subtitle: subtitle != null ? {"en": subtitle} : null,
        contents: {"en": content},
        includePlayerIds: userIds ?? [],
        data: data?.toMap(),
        bigPicture: bigPictureUrl,
        huawei_big_picture: bigPictureUrl,
        android_channel_id: channelId,
        huawei_channel_id: channelId,
      );

      final result = await NUIWeb.get(module: _NUIWEB_ONESIGNAL).doPOST<NUIPushOSResponse>(bodyEnt: notification, headers: _oneSignalPushHeaders(), entity: NUIPushOSResponseEntity());
      if(result.success == true){
        logNUI(_MODULE, "Send push message to users successful");
        return DataResult(success: true, message: "Send push message to users successful");
      }
      else{
        logNUI(_MODULE, "Failed to send push message to users with message: ${result.message}");
        return DataResult(success: false, message: "Failed to send push message to users with message : ${result.message}");
      }
    }catch(e){
      logNUI(_MODULE, "Failed to send push message to users with error : $e");
      return DataResult(success: false, message: "Failed to send push message to users with error: $e");
    }
  }

  @override
  Future<DataResult<List<NUIPushInboxMessage>>> getInboxNotifications({required int page, required int limit}) async{
    final userId = _userId;
    final device = await getDeviceInfo();
    final platform = device?.platform;
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doGET(url: "notification/getNotificationListNew", params: {
      "page": page.toString(),
      "limit": limit.toString(),
      "userId": userId
    }, headers: {
      "Authorization": platform ?? "-",
      "X-DeviceOs": platform ?? "-"
    }, entity: NUIPushStandardResponseEnt<List<NUIPushInboxMessage>>(NUIEntType.LIST_OBJECT, (data) => asListOfObjects(NUIPushInboxMessageEntity(), data)));

    if(result.success == true){
      logNUI(_MODULE, "Get push messages successful");
      return DataResult<List<NUIPushInboxMessage>>(success: true, message: result.message, data: result.body.data);
    }
    else{
      logNUI(_MODULE, "Failed to get push messages");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult<int>> getInboxNotificationCount() async{
    final userId = _userId;
    final device = await getDeviceInfo();
    final platform = device?.platform;
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doGET(url: "notification/getNotificationCount", params: {
      "userId": userId
    }, headers: {
      "Authorization": platform ?? "-",
      "X-DeviceOs": platform ?? "-"
    }, entity: NUIPushStandardResponseEnt<NUINotificationCountPayload>(NUIEntType.OBJECT, (data) => asObject(NUINotificationCountPayloadEntity(), data)));

    if(result.success == true){
      logNUI(_MODULE, "Get notification count successful");
      return DataResult<int>(success: true, message: result.message, data: result.body.data?.notificationCount ?? 0);
    }
    else{
      logNUI(_MODULE, "Failed to get notification count");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult> setInboxNotificationRead({List<String>? notificationId, bool isRead = true}) async{
    final userId = _userId;
    final device = await getDeviceInfo();
    final platform = device?.platform;
    final req = NUINotificationReadPayload(notifId: notificationId ?? [], isRead: isRead);
    final result = await NUIWeb.get(module: NUIWEB_MODULE).doPOST(url: "notification/setNotificationRead", params: {
      "userId": userId
    }, headers: {
      "Authorization": platform ?? "-",
      "X-DeviceOs": platform ?? "-"
    }, bodyEnt: req, entity: NUIPushStandardResponseEnt(NUIEntType.OBJECT, (data) => data));

    if(result.success == true){
      logNUI(_MODULE, "Set notification read successful");
      return DataResult(success: true, message: result.message);
    }
    else{
      logNUI(_MODULE, "Failed to set notification read");
      return DataResult(success: false, message: result.message);
    }
  }
}

class NUIPushBuilder{
  late String _baseUrl;
  String? _oneSignalAuthKey;
  late String _oneSignalAppId;
  late String _userId;
  NUIWebBuilder? _overwriteWebBuilder;
  NUIPushNotificationListener? _notificationReceivedListener;
  NUIPushNotificationListener? _notificationOpenedListener;
  NUIPushUserIDListener? _notificationUserIDListener;

  NUIPushBuilder baseUrl(String baseUrl){
    _baseUrl = baseUrl;
    return this;
  }

  NUIPushBuilder oneSignalAuthKey(String authKey){
    _oneSignalAuthKey = authKey;
    return this;
  }

  NUIPushBuilder userIDChangedListener(NUIPushUserIDListener listener){
    _notificationUserIDListener = listener;
    return this;
  }

  NUIPushBuilder userID(String userId){
    _userId = userId;
    return this;
  }

  /// The module will be replaced with [NUIPush.NUIWEB_MODULE]
  /// The baseURL will be replaced with [_baseUrl]
  NUIPushBuilder overwriteWebBuilder(NUIWebBuilder webBuilder){
    _overwriteWebBuilder = webBuilder;
    return this;
  }

  NUIPushBuilder oneSignalAppId(String appId){
    _oneSignalAppId = appId;
    return this;
  }

  NUIPushBuilder notificationReceivedListener(NUIPushNotificationListener listener){
    _notificationReceivedListener = listener;
    return this;
  }

  NUIPushBuilder notificationOpenedListener(NUIPushNotificationListener listener){
    _notificationOpenedListener = listener;
    return this;
  }

  void _updateOSUserIdLocally(NUIPush push, String userId) async{
    push._registerUserIDIfChanged(userId);
    push._saveUserID(userId);
  }

  Future<NUIPush> build() async{
    //Build shared pref
    await NUISharedPrefBuilder().build();

    NUIWebBuilder()
      .headers(() => {
        "Content-Type": "application/json"
      })
      .module(NUIPush._NUIWEB_ONESIGNAL)
      .baseURL(NUIPush._ONE_SIGNAL_REST_URL)
      .build();

    Map<String, String> pushDefaultHeaders = {"Content-Type": "application/json"};
    if(_overwriteWebBuilder == null){
      NUIWebBuilder()
          .module(NUIPush.NUIWEB_MODULE)
          .headers(() => pushDefaultHeaders)
          .baseURL(_baseUrl)
          .build();
    }else{
      _overwriteWebBuilder!.module(NUIPush.NUIWEB_MODULE);
      _overwriteWebBuilder!.baseURL(_baseUrl);
      _overwriteWebBuilder!.build();
    }

    final push = NUIPush._builder(this);

    OneSignal.shared.setLogLevel(OSLogLevel.verbose, OSLogLevel.none);

    OneSignal.shared.setAppId(_oneSignalAppId);

    await OneSignal.shared.promptUserForPushNotificationPermission(fallbackToSettings: true);

    OneSignal.shared.setNotificationWillShowInForegroundHandler((OSNotificationReceivedEvent event) {
      print("NUIPush: OneSignal received listener for notification will show in foreground");
      // will be called whenever a notification is received
      if(_notificationReceivedListener != null) {_notificationReceivedListener!(NUIPushNotification(osNotification: event.notification));};
    });

    OneSignal.shared.setNotificationOpenedHandler((OSNotificationOpenedResult result) {
      print("NUIPush: OneSignal received listener for notification opened");
      // will be called whenever a notification is opened/button pressed.
      if(_notificationOpenedListener != null) {_notificationOpenedListener!(NUIPushNotification(osNotification: result.notification));};
    });

    OneSignal.shared.setPermissionObserver((OSPermissionStateChanges changes) {
      // will be called whenever the permission changes
      // (ie. user taps Allow on the permission prompt in iOS)
    });

    OneSignal.shared.setSubscriptionObserver((OSSubscriptionStateChanges changes) {
      // will be called whenever the subscription changes
      //(ie. user gets registered with OneSignal and gets a user ID)
      print("NUIPush: OneSignal subscription observer received new changes, from.userId=${changes.from?.userId}, to.userId=${changes.to?.userId}");
      if(!isNullOrEmpty(changes.to.userId)){
        _updateOSUserIdLocally(push, changes.to.userId!);
        if(_notificationUserIDListener != null) {_notificationUserIDListener!(changes.to.userId!);}
      }
      // else{
      //   _updateOSUserIdLocally(push, changes.from.userId!);
      //   if(_notificationUserIDListener != null) {_notificationUserIDListener!(changes.to.userId);}
      // }
    });

    OneSignal.shared.setEmailSubscriptionObserver((OSEmailSubscriptionStateChanges emailChanges) {
      // will be called whenever then user's email subscription changes
      // (ie. OneSignal.setEmail(email) is called and the user gets registered
    });

    final deviceState = await OneSignal.shared.getDeviceState();
    final deviceToken = deviceState?.userId;
    if(!isNullOrEmpty(deviceToken)){
      _updateOSUserIdLocally(push, deviceToken!);
      if(_notificationUserIDListener != null) {_notificationUserIDListener!(deviceToken);}
    }

    return push;
  }
}